open Evt
open List

(* this module dumps an event stream to stdout, conservatively. *)

let otag t = 
  print_char '<'; 
  print_string t; 
  print_char '>'

let ctag t = 
  print_char '<'; 
  print_char '/'; 
  print_string t; 
  print_char '>'

let prop_2_str = function
    B -> "B"
  | EM -> "EM"
  | I -> "I"
  | PL -> "PL"
  | S -> "S"
  | TT -> "TT"
  | U -> "U"
  | Sz x -> (string_of_int x)
  | Col Red -> "r"
  | Col Green -> "g"
  | Col Blue -> "b"
  | Col Cyan -> "c"
  | Col Magenta -> "m"
  | Col Yellow -> "y"
  | Col Black -> "k"
  | Col White -> "w"
  | Dummy -> "DUMMY"

let enter p = match p with Dummy -> () | _ -> otag (prop_2_str p)
let leave p = match p with Dummy -> () | _ -> ctag (prop_2_str p)  

(* the first section deals with constructing conservative difference
   lists of existing property lists.  the property lists are _ordered_,
   such that we cannot in general arrange to leave a tag before one of
   its enclosed tags, etc. special cases where we might be able to do so
   are dealt with later on *)
	      
let diff (a,b) =
  let rec chomp pair = 
    match pair with
	([],x) -> ([],x)
      | (x,[]) -> (x,[])
      | (x::xs,y::ys) when x=y -> chomp (xs,ys)
      | (x,y) -> (x,y)
  in
    chomp (a, b)       

let full_diff (a,b) =
  let not_dummy = fun x -> (match x with Dummy -> false | _ -> true) in
  let (a0,b0) = (filter not_dummy a, filter not_dummy b) in
  let (a1,b1) = diff (a,b) in
  let (a2,b2) = diff (rev a1, rev b1) in
    (rev a2, b2)


(* the next section deals with computing whether one property list can
   be said to be "extending" another. the fix_exts routine returns a new
   "current" property list (as well as the preferred leave/enter diff
   pair) by calculating whether it ought to consider itself an extension
   of the previous list, or a deletion & possible reconstruction. this
   addresses the problem of merging neighbouring mutually exclusive
   extents. *)


let extends pair =
  match pair with
      (Col x, Col y) -> (x != y)
    | (Sz x, Sz y) -> (x != y)
    | _ -> false
	
let rec ext pair =  
  match pair with
    | (a::az,b::bz) when a=b -> b :: (ext (az,bz))
    | (a::az,b::bz) when extends (a,b) -> a :: (ext (az,b::bz))
    | (_,b) -> b
	
let fix_exts (last,curr) =
  let res = rev (ext (rev last,rev curr)) in
  let (leave1,enter1) = full_diff (last,curr) in
  let (leave2,enter2) = full_diff (last,res) in
    if (length leave1 + length enter1 >
	length leave2 + length enter2)
    then (res,(leave2,enter2))
    else (curr,(leave1,enter1))


(* the next section deals with normalizing a property list. we do this
   to both our "leave" and "enter" property lists before outputting them.
   this removes some duplication of property-setting. it is accomplished
   via a fold which accumulates the head state and the minimal property
   list required to achieve said state. *)

type propset = { b  : bool; em : bool;
		 i  : bool; s  : bool;
		 tt : bool; u  : int;
		 size : int; 
		 color : color }

let plain = { b=false; em=false; i=false; 
	      s=false; tt=false; u=0; 
	      size=0; color=Black }

let is_plain p = (p.u = 0 && not (p.s || p.em || p.i || p.b || p.tt))
    
let set_prop (ls,pset) prop = 
  let rec plain_nuke list = match list with
      (p::pz) when mem p [B;I;TT;U;S] -> plain_nuke pz
    | other -> other
  in
  match prop with
      Col c when pset.color != c -> (prop::ls, {pset with color=c})
    | Sz x  when pset.size  != x -> (prop::ls, {pset with size=x})
    | B     when not pset.b      -> (prop::ls, {pset with b=true})
    | I     when not pset.i      -> (prop::ls, {pset with i=true})
    | TT    when not pset.tt     -> (prop::ls, {pset with tt=true})
    | EM    when not pset.s      -> (prop::ls, {pset with em=(not pset.em)})
    | S     when not pset.s      -> (prop::ls, {pset with s=true; em=false})
    | U     when pset.u < 3      -> (prop::ls, {pset with u=(pset.u+1)})
    | PL    when not (is_plain pset) 
	-> (prop::(plain_nuke ls), {plain with 
				      size=pset.size; 
				      color=pset.color})
    | _ -> (Dummy::ls,pset)

let normalize ls = fold_left set_prop ([],plain) (rev ls)


(* whitespace elimination.. this bit is a little "imperative" for a
   functional programming competition; but then, this is a multi-paradigm
   language. *)

let in_ws = ref false
let last_ws_ctx = ref plain
let string_out (str,state) =
  if state.tt 
  then str
  else 
    let nstr = String.create (String.length str) in
    let n = ref 0 in
    let ws_ctx = {plain with u=state.u; 
		    color=state.color; 
		    size=state.size} 
    in
    for i = 0 to String.length str - 1 do
      let ch = str.[i] in
	if ch = ' ' || ch = '\t' || ch = '\r' || ch = '\n' 
	then 
	  begin
	    if !in_ws && (ws_ctx.u = 0 || (ws_ctx = !last_ws_ctx))
	    then () 
	    else 
	      begin 
		nstr.[!n] <- ' '; 
		n := !n + 1;
		in_ws := true; 
		last_ws_ctx := ws_ctx
	      end
	  end
	else (in_ws := false; nstr.[!n] <- ch; n := !n + 1)
    done;
      String.sub nstr 0 !n
      
(* finally, we have our lexeme event handler, which takes an output
   state and an event, and pumps the event either into the state or out
   to stdout *)

let dump_event (a,b) ev = match ev with
    Push p -> (a,p::b)
  | Pop -> (a,tl b)
  | Txt t -> 	  
      let (a',_) = normalize a in
      let (b',s) = normalize b in
      let str = string_out (t, s) in
	if String.length str = 0 
	then (a,b)  (* "don't flush output yet" *)
	else
	  let (b'',(left,entered)) = fix_exts (a',b') in
  	    iter leave left;
  	    iter enter entered;
  	    print_string str;
	    (b'',b)

let lexbuf = (Lexing.from_channel stdin)

let _ = 
  let rec opt (a,b) =    
    match Lex.token lexbuf with
	None -> iter leave a
      | Some ev -> 
	  let st2 = dump_event (a,b) ev in
	    opt st2
  in
    opt ([],[])

